#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>

#include "ptopcode.h"
#include "vcc.h"

extern int cstacktop;
extern int savedcstacktop;
extern jmp_buf topenv;

#define EVALSTACKLEN 200
#define CALLEDLEN 100
#define LABELSTACKLEN 16

#define isclass(c) (c == CLS)
#define isobject(c) (c > CLS)

t_vardat frames[128];

int frametop;
int savedframetop;

int framestack[EVALSTACKLEN]; /* 2018 */
int lvarstack[EVALSTACKLEN]; /* 2018 */
int lvarp;

struct {
  int labname;
  jmp_buf jump;
} labelstack[LABELSTACKLEN];

jmp_buf called[CALLEDLEN];
int calledtop;

t_vardat *thisptr;

void *cfunsearch();
int *listaccess();
t_vardat *createobj(int);
t_vardat *objaccess(int, int *);

int evalinit()
{
  calledtop = 0;
  lvarp = 0;
  frametop = 0;
  savedframetop = 0;
  thisptr = NULL;
}

int methodsearch(t_vardat *objp, int mname)
{
  int i;
  t_vardat *varp = objp->val;

  for (i = 0; i < objp->paramlen; i++) {
    if (mname == varp[i].name) {
      if (debug) 
	printf("methodsearch for %s.%s: i = %d\n", namepool+(objp->kind), namepool+mname, i);
      return i;
    }
  }
    return -1;
}

int varsearch(int varname)
{
  int i;

  if (savedframetop == 0) {
    if (debug)
      printf("varsearch: global for %s\n", namepool+varname);
    return -1;
  }
  for (i = frametop-1; i >= framestack[savedframetop-1]; i--) {
    if (varname == frames[i].name) {
      if (debug)
	printf("varsearch: local for %s (varpos = %d)\n", namepool+varname, i);
      return i;
    }
  }
  if (debug)
    printf("varsearch: *global for %s\n", namepool+varname);
  return -1;
}

int globalvarsearch(int varname)
{
  ENTRY e, *r;
  e.key = namepool+varname;
  hsearch_r(e, FIND, &r, &varhash);
  if (r == NULL) {
    int j;

    j = varstop++;
    vars[j].kind = VAL;
    vars[j].name = varname;
    vars[j].val = 0;

    e.data = (void *)j;
    hsearch_r(e, ENTER, &r, &varhash);
    return j;
  } else {
    return (int)(r->data);
  }
}

/* 2018 */
int set(int lhs, int val, int kd)
{
  int srcpos, dstpos, res, methodkind=-1;
  t_vardat *src, *dst;

  switch (kd) {
  case VAL:
  case LST:
    res = eval(val);
    break;
  case CLS:
    src = createobj(val);
    res = src;
    break;
  }

  if (cstack[lhs].opcode == LISTREF) {
    int *ldst = listaccess(lhs);
    *ldst = res;
    return res;
  } else {
    int varname = cstack[lhs].val1;

    switch (cstack[lhs].opcode) {
    case VAR:
      if ((dstpos = varsearch(varname)) == -1) { /* set to global var */
	dst = &(vars[globalvarsearch(varname)]);
      } else {
	dst = &(frames[dstpos]);
      }
      break;
    case OBJREF:
      dst = objaccess(lhs, &methodkind);
      break;
    }
    if (methodkind ==LST) {
      int *ldst = dst;
      *ldst = res;
      return res;
    }
  }

  if (cstack[val].opcode == VAR) {
      if ((srcpos = varsearch(cstack[val].val1)) == -1) {
	src = &(vars[globalvarsearch(cstack[val].val1)]);
      } else {
	src = &(frames[srcpos]);
      }
      dst->kind = src->kind;
      dst->val = src->val;
      dst->param = src->param;
      dst->paramlen = src->paramlen;
  } else if (kd == CLS) {
      dst->kind = src->kind;
      dst->val = src->val;
      dst->paramlen = src->paramlen;
  } else { /* also applies to LST. */
    dst->kind = kd;
    dst->val = res;
  }
  return dst->val;
}

cstacksave()
{
  savedcstacktop = cstacktop;
}

cstackclear()
{
	cstacktop = savedcstacktop;
}

static int labelstacktop = -1;

int dolabel(int labelname)
{
  labelstack[++labelstacktop].labname = labelname;
  if (debug) {
    int i;
    for (i = 0; i <= labelstacktop; i++) {
      printf("label %d: %s\n", i, namepool+labelstack[i].labname);
    }
  }
  return labelstacktop;
}

int findlabel(int labelname)
{
  if (labelname == 0) { /* default */
    return labelstacktop;
  } else {
    int i;
    for (i = labelstacktop; i >=0; i--) {
      if (labelstack[i].labname == labelname)
	return i;
    }
    /* return labelstacktop; */
    {char msg[256];
      sprintf(msg, "%s: label out of scope", namepool+labelname);
      yyerror(msg);
      return -1;
    }

  }
}
void rewindlabel(int labelname, int breaking)
{
  if (labelname == 0) { /* default */
    if (labelstacktop >= 0)
      if (breaking)
	labelstacktop--;
  } else {
    int i;
    for (i = labelstacktop; i >=0; i--) {
      if (labelstack[i].labname == labelname) {
	if (breaking) labelstacktop = i-1;
	else labelstacktop = i;
	break;
      }
    }
  }
}

eval(int top)
{
  int res;
  int savedcalledtop;
  t_vardat *savedthisptr;

  switch (cstack[top].opcode) {
  case LIT: return cstack[top].val1;
  case VAR:  /* 2018 */
    {int varpos;
      if ((varpos = varsearch(cstack[top].val1)) == -1) {
	return vars[globalvarsearch(cstack[top].val1)].val;
      } else {
	return frames[varpos].val;
      }
    }
  case ADD: 
    return eval(cstack[top].val1)+eval(cstack[top].val2);
  case SUB: 
    return eval(cstack[top].val1)-eval(cstack[top].val2);
  case MUL: 
    return eval(cstack[top].val1)*eval(cstack[top].val2);
  case MOV:
    return set(cstack[top].val1, cstack[top].val2, VAL);
    /* 2018 */
  case GREATER:
    return (eval(cstack[top].val1) > eval(cstack[top].val2));
  case GREATEREQ:
    return (eval(cstack[top].val1) >= eval(cstack[top].val2));
  case LESSEQ:
    return (eval(cstack[top].val1) <= eval(cstack[top].val2));
  case LESS:
    return (eval(cstack[top].val1) < eval(cstack[top].val2));
  case EQUAL:
    return (eval(cstack[top].val1) == eval(cstack[top].val2));
  case NOTEQUAL:
    return (eval(cstack[top].val1) != eval(cstack[top].val2));
  case IF:
    return ((eval(cstack[top].val1)? eval(cstack[cstack[top].val2].val1): 
	     eval(cstack[cstack[top].val2].val2)));
  case COMPOUND:
    eval (cstack[top].val1); 
    return eval(cstack[top].val2);
  case WHILE:
    {
      int ltop, res, lab, wcond, wbody;
      lab = cstack[top].val1;
      wcond = cstack[cstack[top].val2].val1;
      wbody = cstack[cstack[top].val2].val2;
      ltop = dolabel(lab);
      if (debug) {
	printf("lab = %d, ltop = %d\n", lab, ltop);
      }
      res = setjmp(labelstack[ltop].jump);
      switch (res) {
      case 0: /* setjump enter */ break;
      case BREAK: {
	if (debug) {
	  printf("break jump from %s\n", namepool+lab);
	}
	goto breaklabel;
      }
	case REDO: {
	  if (debug) {
	    printf("redo jump from %s\n", namepool+lab);
	  }
	    rewindlabel(lab, 0); 
	    break;
	}
      }
      while (eval(wcond)) eval(wbody);
    breaklabel:
      rewindlabel(lab, 1);
      return 1;
    }
  case FUNCALL:
    { int i;
      t_vardat *v;

      if (thisptr != NULL) { /* within an object */
	i = methodsearch(thisptr, cstack[top].val1);
	if (debug)
	  printf("method %d on %d\n", i, thisptr);
	if (i < 0) return 0;
	v = (t_vardat *)(thisptr->val)+i;
      } else {
	if ((i = varsearch(cstack[top].val1)) == -1) {
	  i = globalvarsearch(cstack[top].val1);
	  v = vars+i;
	} else {
	  v = frames+i;
	}
      }
      if (v->val == 0) {
	void (*cfun)();

	/* undefined */
	if ((cfun = cfunsearch(cstack[top].val1)) != NULL) {
	  v->kind = CFUN;
	  v->val = (void *)cfun;
	}
      }
      switch (v->kind) {
      case VAL:
	return 0;
      case FUN:
	savedcalledtop = calledtop;
	res = setjmp(called[calledtop++]);
	if (res == 0) {
	  int r;

	  funprolog(v->param, cstack[top].val2, v->paramlen);
	  r = eval(v->val);
	  calledtop = savedcalledtop;
	  funepilog();
	  return r;
	  
	} else {
	  int r;
	
	  r = eval(res);
	  funepilog();
	  return r;
	}
	break;
      case CFUN:
	return cfuncall(v->val, cstack[top].val2);
      default:
	fprintf(stderr, "FATAL in VAR (kind = %d)\n", vars[cstack[top].val1].kind);
	exit(1);
      }
    }
  case FMOV:
    {int i;
      i = globalvarsearch(cstack[top].val1);

    vars[i].kind = FUN;
    vars[i].val = cstack[top].val2;
    vars[i].param = vars[cstack[top].val2].param;
    return (vars[i].val);
    }
  case MYVAR:
    return lvarpush(cstack[top].val1);
  case LENV:
    { int r;
      lvarstack[lvarp++] = frametop;
      r = eval(cstack[top].val1);
      frametop = lvarstack[--lvarp];
      return r;
    }
  case NEG:
    return !eval(cstack[top].val1);
  case LISTDEF:
    return set(cstack[top].val1, cstack[top].val2, LST);
  case LISTA:
    return listalloc(top);
  case LISTREF:
    return *(listaccess(top));
  case RET:
    /*     fprintf(stderr, "returning to top with %d\n", cstack[top].val1); */
      longjmp(called[--calledtop], cstack[top].val1);
      /* We use the fact that in RET X, X is never 0 */
    break;
  case BREAK: {
    int lab = findlabel(cstack[top].val1);
    if (lab <0) longjmp(topenv, -1);
    if (debug) printf("break jump to %s\n", namepool+labelstack[lab].labname);
      longjmp(labelstack[lab].jump, BREAK);
    break;
  }
  case REDO: 
    {
      int lab = findlabel(cstack[top].val1);
      if (lab <0) longjmp(topenv, -1);
      if (debug) printf("redo jump to %s\n", namepool+labelstack[lab].labname);
      longjmp(labelstack[lab].jump, REDO);
      break;
    }
  case NEWOBJ:
    return set(cstack[top].val1, cstack[top].val2, CLS);
  case THIS:
    return (int)thisptr;
  case OBJREF:
    {
      int i, methodkind;
      t_vardat *oref;

      savedthisptr = thisptr;
      oref = objaccess(top, &methodkind);
      thisptr = savedthisptr;
      switch (methodkind) {
      case VAL:
	return oref->val;
      case CLS:
      case FUN:
	return (int)oref;
      case LST:
	return *((int *)oref);
      }
    }
  }
}

lvarpush(int plist)
{
  for (; cstack[plist].opcode == PARAM; plist = cstack[plist].val2, frametop++) {
    frames[frametop].name = cstack[cstack[plist].val1].val1;
    frames[frametop].kind = VAL;
    frames[frametop].val = 0;
    if (debug) {
      printf("setting local var %c with %d\n", frames[frametop].name, frametop);
    }
  }
}

funprolog(int paramlist, int arglist, int plen)
{
  /* 2018 */
  funprolog1(paramlist, arglist, plen);
  framestack[savedframetop++] = frametop;
  frametop += plen;
}

int setparamarg(int plist, int alist, int framebase, int i)
{
  int varpos, name;

  name = frames[framebase+i].name = cstack[plist].val1;

  if (cstack[alist].opcode == VAR) {
    /* argument is also a variable */
    if ((varpos = varsearch(cstack[alist].val1)) == -1) { 
      /* arg. is also global */
      frames[framebase+i] = vars[globalvarsearch(cstack[alist].val1)];
    } else {
      frames[framebase+i] = frames[varpos];
    }
  } else if (cstack[alist].opcode == THIS) {
    frames[framebase+i].kind = thisptr->kind;
    frames[framebase+i].val = thisptr->val;
    frames[framebase+i].paramlen = thisptr->paramlen;
  } else {
    frames[framebase+i].kind = VAL;
    frames[framebase+i].val = eval(alist);
    if (debug)
      printf("setparamarg: setting %d to %s\n", frames[framebase+i].val, namepool+frames[framebase+i].name);
  } 
  return i;
}

int funprolog1(int paramlist, int arglist, int plen)
{
  int i;

  for (i = 0; i < plen; i++, paramlist = cstack[paramlist].val2, arglist = cstack[arglist].val2) {
    if (cstack[paramlist].opcode == VAR) {
      setparamarg(paramlist, arglist, frametop, i);
      break;
    } else if (cstack[paramlist].opcode == PARAM) {
      /*       if (cstack[paramlist].val1 == 0) return 0; */
      setparamarg(cstack[paramlist].val1, cstack[arglist].val1, frametop, i);
    }
  }
  return i;
}

funepilog()
{
  frametop = framestack[--savedframetop];
}

cfuncall(int (*f)(), int args)
{
#define MAXARG 8

  int i;
  int alist[MAXARG];

  for (i = 0; ; i++) {
    if (cstack[args].opcode == ARG) {
      alist[i] = eval(cstack[args].val1);
    args = cstack[args].val2;
    } else
      break;
  }
  return f(i, alist);
}


listalloc(int top)
{
  int li = top, size, i;
  int *p;

  size = eval(cstack[li].val1);
  p = malloc(size * sizeof(int));
  if (cstack[li].val2 != 0) {
    for (i = 0; i < size; i++) {
      p[i] = listalloc(cstack[li].val2);
    }
  }
  return (int)p;
}


int *listaccess(int top)
{
  int varpos, *sublist, offset, i;

  if (cstack[cstack[top].val1].opcode == VAR) {
    int var = cstack[top].val1;

    if (thisptr != NULL) { /* object reference */
      i = methodsearch(thisptr, cstack[var].val1);
      if (i >=0) 
	sublist = ((t_vardat *)(thisptr->val))[i].val;
      else {
	fprintf(stderr, "%s is not a method\n", namepool+cstack[var].val1);
	sublist = NULL;
      }
    } else if ((varpos = varsearch(cstack[var].val1)) == -1) {
      varpos = globalvarsearch(cstack[var].val1);

      if (vars[varpos].kind != LST) return 0;
      sublist = (int *)vars[varpos].val;
    } else { 
      if (frames[varpos].kind != LST) return 0;
      sublist = (int *)frames[varpos].val;
    }
  } else { /* LISTREF */
    sublist = listaccess(cstack[top].val1);
  }
  offset = eval(cstack[top].val2);
  return sublist + offset;
}

t_vardat *createobj(int top)
{

  int i, varp = globalvarsearch(cstack[top].val1);
  int inimethod;
  t_vardat *re, *newobj;

  if (!isclass(vars[varp].kind)) {
    char msg[256];
    sprintf(msg, "%s is not a class\n", namepool+cstack[top].val1);
    yyerror(msg);
    return NULL;
  }

  re = malloc(sizeof(t_vardat));
  re->kind = cstack[top].val1; /* class name of the object */
  re->val = newobj = malloc(sizeof (t_vardat) * vars[varp].paramlen);
  re->paramlen = vars[varp].paramlen;

  for (i = 0; i < vars[varp].paramlen; i++) {
    newobj[i].kind = ((t_vardat *)vars[varp].val)[i].kind;
    newobj[i].name = ((t_vardat *)vars[varp].val)[i].name;
    if (newobj[i].kind == FUN) {
      newobj[i].val =  ((t_vardat *)vars[varp].val)[i].val;
      newobj[i].param = ((t_vardat *)vars[varp].val)[i].param;
      newobj[i].paramlen = ((t_vardat *)vars[varp].val)[i].paramlen;
      newobj[i].res_t = VAL;
    }
  }

  inimethod = methodsearch(re, initpos());

  if (inimethod>=0) {
    t_vardat *savedthisptr;

    savedthisptr = thisptr;
    thisptr = re;
    cstack[top].opcode = FUNCALL; /* borrow from funcall */
    cstack[top].val1 = initpos();

    eval(top);
    thisptr = savedthisptr;
  }
  return re;
}


t_vardat *objaccess(int top, int *methodkind)
{
  int i, varpos;
  t_vardat *varp, *meth;

  if (cstack[top].opcode == THIS) {
    varp = frames + 0; /* "this" is the first parameter */
    thisptr = varp;
    *methodkind = CLS;
    return varp;
  } else if (cstack[top].opcode == VAR) {
    if ((varpos = varsearch(cstack[top].val1)) == -1) {
      varpos = globalvarsearch(cstack[top].val1);
      varp = vars + varpos;
    } else {
      varp = frames + varpos;
    }
    thisptr = varp;

    *methodkind = CLS;
    return varp;
  } else { /* OBJREF */
    int i, mk;
    t_vardat *subobj;

    subobj = (t_vardat *)objaccess(cstack[top].val1, &mk);
    if (!isobject(subobj->kind)) {
      fprintf(stderr, "%s is not an object\n", namepool+subobj->name);
      varp = NULL;
      *methodkind = -1;
    } else if (cstack[cstack[top].val2].opcode == VAR) {
      i = methodsearch(subobj, cstack[cstack[top].val2].val1);
      if (i >=0) 
	varp = ((t_vardat *)(subobj->val))+i;
      else {
	fprintf(stderr, "%s is not a method\n", namepool+cstack[cstack[top].val2].val1);
	varp = NULL;
      }
      *methodkind = VAL;
    } else if (cstack[cstack[top].val2].opcode == FUNCALL) {
      varp = subobj = eval(cstack[top].val2);
      *methodkind = FUN;
    } else if (cstack[cstack[top].val2].opcode == LISTREF) {
      /*       varp = subobj = eval(cstack[top].val2); */
      varp = subobj = listaccess(cstack[top].val2);
      *methodkind = LST;
    }
    thisptr = varp;
    return varp;
  }
}

