%{
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "ptopcode.h"

extern int debug;
extern int optimize;
extern int codepc;
extern int savedcodepc;
%}
%token DIGIT
%token VARIABLE
%token GE
%token GT
%token LE
%token LT
%token EQ
%token NE
%token NOT
%token ASSIGN
%token WH
%token DE
%token RE
%token MY
%token IFIF
%token ELSE
%token FOR
%token IMPORT
%token LIST
%token BRK
%token RDO
%token NEW
%token CLASS
%token VL
%token THS
%token OBJINIT

%left ASSIGN
%left GE GT LE LT EQ NE
%left '+' '-'
%left '*' '/'
%right UMINUS
%%

lines : lines line
	| line
	|
	;
line : expr ';'	{int i, pt;
	         savedcodepc = codepc;
		 pt = $1;
		 if (optimize) treeopt(pt);
		 codemit(pt); 
		 printemit();
		 if (debug) disassemble();
		 vmexec(savedcodepc);
	         codepc = savedcodepc;
	        }
	| defun ';' {if (debug) disassemble();}
 	| defclass ';' {if (debug) disassemble();}
	| importdl ';'
	| error ';'	{yyerror(""); yyerrok;}
	;
compound: compound ';' expr 	{$$ = push(COMPOUND, $1, $3);}
	| expr
	;
aexpr: 	expr '+' expr	{$$ = push(ADD, $1, $3);}
	| expr '-' expr	{$$ = push(SUB, $1, $3);}
	| expr '*' expr	{$$ = push(MUL, $1, $3);}
	| '(' expr ')'	{$$ = $2;}
	| '-' expr %prec UMINUS	{int x; x = push(LIT, -1, 0); $$=push(MUL, $2, x);}
	;
condexpr: expr GE expr	{$$ = push(GREATEREQ, $1, $3);}
	| expr GT expr	{$$ = push(GREATER, $1, $3);}
	| expr LE expr	{$$ = push(LESSEQ, $1, $3);}
	| expr LT expr	{$$ = push(LESS, $1, $3);}
	| expr EQ expr	{$$ = push(EQUAL, $1, $3);}
	| expr NE expr	{$$ = push(NOTEQUAL, $1, $3);}
	| NOT expr {$$ = push(NEG, $2, 0);}
	;
lhs: VARIABLE {$$=push(VAR, $1, 0);}
	| objref 
	| listref
	;
assignment: lhs ASSIGN expr {$$ = push(MOV, $1, $3);}
	| lhs ASSIGN newobject {$$ = push(NEWOBJ, $1, $3);}
	| lhs ASSIGN listdef {$$ = push(LISTDEF, $1, $3);}
	;
expr : assignment
	| ifstatement
	| loopstatement
	| '{' compound '}'	{int x; x = $2; $$=push(LENV, x, 0);}
	| aexpr
	| condexpr
	| RE expr	{$$ = push(RET, $2, 0);}
	| MY '(' varlist ')' {$$ = push(MYVAR, $3, 0);}
	|DIGIT		{$$ = push(LIT, yylval, 0);}
	|VARIABLE	{$$ = push(VAR, yylval, 0);}
	|VARIABLE '(' exprlist ')'	{$$ = push(FUNCALL, $1, $3);}
	| listref
	| objref
	|{$$=push(LIT, 0, 0);}
	;
ifstatement: '(' expr ')' '?' expr ':' expr	{int x; x = push(ALT, $5, $7); $$ = push(IF, $2, x);}
	|IFIF '(' expr ')' expr {int x, z; z = push(LIT, 0, 0); x = push(ALT, $5, z); $$ = push(IF, $3, x);}
	| IFIF '(' expr ')' expr ELSE expr {int x; x = push(ALT, $5, $7); $$ = push(IF, $3, x);}
	;
loopstatement: WH '(' expr ')' expr {int z = push(ALT, $3, $5); $$ = push(WHILE, 0, z);}
	| VARIABLE ':' WH '(' expr ')' expr {int z = push(ALT, $5, $7); $$ = push(WHILE, $1, z);}
	| BRK {$$=push(BREAK, 0, 0);}
	| RDO {$$=push(REDO, 0, 0);}
	| BRK VARIABLE {$$=push(BREAK, $2, 0);}
	| RDO VARIABLE {$$=push(REDO, $2, 0);}
	;
exprlist: expr ',' exprlist	{$$ = push(ARG, $1, $3);}
	| expr 			{$$ = push(ARG, $1, 0);}
	|			{$$ = 0;}
	;
varlist: VARIABLE ',' varlist	{int x = push(VAR, $1, 0); $$ = push(PARAM, x, $3);}
	| VARIABLE 		{int x = push(VAR, $1, 0); $$ = push(PARAM, x, 0);}
	|			{$$ = 0;}
	;
methoddef: VL VARIABLE {$$=push(VAR, $2, 0);}
	| VL VARIABLE '(' varlist ')' expr {int x, y, z; x = push(VAR, thispos(), 0); z = push(PARAM, x, $4); y = push(PARAM, z, $6); $$=push(FUNCALL, $2, y);} /* the first parameter is always "this" */
	| OBJINIT '(' varlist ')' expr {int x, y, z; x = push(VAR, thispos(), 0); z = push(PARAM, x, $3); y = push(PARAM, z, $5); $$=push(FUNCALL, initpos(), y);} /* the first parameter is always "this" */
	| VARIABLE VARIABLE  {$$=push(OBJREF, $2, $1);}
	| VARIABLE VARIABLE '(' varlist ')' expr {int x, y, z; x = push(VAR, thispos(), 0); z = push(PARAM, x, $4); y = push(PARAM, z, $6); $$=push(FUNCALL, $2, y);} /* the first parameter is always "this" */
	;
methodlist: methoddef ';' methodlist {$$ = push(PARAM, $1, $3);}
	| methoddef {$$=push(PARAM, $1, 0);}
	| {$$=0;}
	;
newobject: NEW VARIABLE '(' exprlist ')' {int x, y; x = push(THIS, 0, 0); y = push(ARG, x, $4); $$=push(NEWOBJ, $2, y);}
	;
objref: objref '.' VARIABLE {int x = push(VAR, $3, 0); $$=push(OBJREF, $1, x);}
	| objref '.' VARIABLE '(' exprlist ')' {int x, y; y = push(ARG, push(THIS, 0, 0), $5); x = push(FUNCALL, $3, y); $$=push(OBJREF, $1, x);} /* automatically insert "this" at the first position */
	| objref '.' listref {$$=push(OBJREF, $1, $3);} /* automatically insert "this" at the first position */
	| THS {$$=push(VAR, thispos(), 0);}
	| VARIABLE {$$= push(VAR, $1, 0);}
	;

listdef: LIST '(' expr ')' {$$ = push(LISTA, $3, 0);}
	| LIST '(' expr ',' listdef ')' {$$ = push(LISTA, $3, $5);}
	;
listref: VARIABLE '[' expr ']' {int x; x = push(VAR, $1, 0); $$ = push(LISTREF, x, $3);}
	| listref '[' expr ']' {$$ = push(LISTREF, $1, $3);}
	;

defun: DE VARIABLE '(' varlist ')' expr {fundef($2, $4, $6); funcodemit($2);}
defclass: CLASS VARIABLE '{' methodlist '}' {classdef($2, $4); classcodemit($2);}
importdl: IMPORT VARIABLE {importfun($2);}

%%
#include "vcc.h"
#include "vm.h"

struct hsearch_data namhash;
struct hsearch_data varhash;

#define STACKLEN 1000

t_ydat cstack[STACKLEN];

int cstacktop;
int savedcstacktop;

t_vardat vars[128];
int varstop;

char namepool[NAMEPOOLSIZE];
int namepooltop;

#include "lex.c"

yyinit()
{
  cstacktop = 0;
  savedcstacktop = 0;
  varstop = 0;
  namepooltop = 0;

  cstack[cstacktop++].opcode = -1; /* match nothing */
  savedcstacktop = cstacktop;

  memset(&namhash, 0, sizeof(namhash));
  memset(&varhash, 0, sizeof(varhash));
  if (hcreate_r(HSIZE, &namhash) == 0 || hcreate_r(HSIZE, &varhash) == 0) {
    fprintf(stderr, "hash memory for PARSE is not enough\n");
    exit(1);
  }

  regkeywords();
  codemitinit();
  vminit();
}

push(int op, int val1, int val2)
{
  cstack[cstacktop].opcode = op;
  cstack[cstacktop].val1 = val1;
  cstack[cstacktop].val2 = val2;
  return cstacktop++;
}

int plistlen(int paramlist)
{
  int i;

  if (paramlist == 0) return 0;
  for (i = 0; cstack[paramlist].opcode == PARAM; i++) {
    paramlist = cstack[paramlist].val2;
  }

  return i;
}

fundef(int varname, int paramlist, int funbody)
{
  int i;
  ENTRY e, *r;
  e.key = namepool+varname;

  hsearch_r(e, FIND, &r, &varhash);
  if (r == NULL) { 
    /* new entry */
    i = varstop++;
    e.data = (void *)i;
    hsearch_r(e, ENTER, &r, &varhash);
  } else {
    i = (int)(r->data);
  }
  if (debug == 1)
    printf("registering %s to %d\n", namepool+varname, i);

  vars[i].kind = FUN;
  vars[i].name = varname;
  vars[i].val = funbody;
  vars[i].param = paramlist;
  vars[i].paramlen = plistlen(paramlist);   /* 2018 */

  savedcstacktop =funbody+1; 
  /* We use the fact that parameters are pushed before funbody */
  if (optimize) {
    treeopt(funbody);
    funopt(funbody);
  }
  savedcstacktop = cstacktop+1;
  return cstacktop++;
}

int classdef(int cname, int cbody)
{
  int i, plen;
  ENTRY e, *r;
  e.key = namepool+cname;

  hsearch_r(e, FIND, &r, &varhash);
  if (r == NULL) { 
    /* new entry */
    i = varstop++;
    e.data = (void *)i;
    hsearch_r(e, ENTER, &r, &varhash);
  } else {
    i = (int)(r->data);
  }
  if (debug)
    printf("registering %s to %d\n", namepool+cname, i);

  plen = plistlen(cbody);
  if (debug)
    printf("registering %s with %d method(s)\n", namepool+cname, plen);
  vars[i].kind = CLS; 
  vars[i].name = cname;
  vars[i].val = cbody;
  vars[i].paramlen = plen;

  savedcstacktop = cstacktop; 
  return cstacktop;
}
