#include "vcc.h"
#include "ptopcode.h"

#define islit(op) ((op)==LIT)
#define isoperator(op) ((op) >= ADD && (op) <= DIV || (op) >= GREATER && (op) <= NOTEQUAL)
#define compconst(op) {cstack[top].val1 = cstack[left].val1 op cstack[right].val1;}

treeopt(int top)
{
  constfolding(top);
  branchopt(top);
}

funopt(int fbody)
{
  treeopt(fbody);
  valuenumbering(fbody);
}

constfolding(int top)
{
  int left, right;
  int op;

  if (top == 0) return;
  op = cstack[top].opcode;

  if (islit(op)) return;
  trivialization(top);
  if (isoperator(op)) {
    left = cstack[top].val1;
    right = cstack[top].val2;
    constfolding(left); 
    constfolding(right);
    trivialization(top);
    if (islit(cstack[left].opcode) && islit(cstack[right].opcode)) {
      cstack[top].opcode = LIT;
      switch (op) {
      case ADD: compconst(+);break;
      case SUB: compconst(-);break;
      case MUL: compconst(*);break;
      case DIV: compconst(/);break;
      case GREATER: compconst(>);break;
      case GREATEREQ: compconst(>=);break;
      case LESSEQ: compconst(<=);break;
      case LESS: compconst(<);break;
      case EQUAL: compconst(==);break;
      }
    }
  } else {
    switch (op) {
    case MOV: 
    case FMOV:
    case FUNCALL:
      constfolding(cstack[top].val2); break;
    case RET:
    case LENV:
      constfolding(cstack[top].val1); break;
    case IF: 
    case ALT:
    case COMPOUND:
    case WHILE:
    case ARG:
      constfolding(cstack[top].val1); constfolding(cstack[top].val2);
    default:
      break;
    }
  }
  doassoc(top);
// Associativity
}

int trivialization(int top)
{
  int op, op1, op2;

  op = cstack[top].opcode;

  if (op == ADD) {
    op1 = cstack[top].val1;
    op2 = cstack[top].val2;

    if (islit(cstack[op1].opcode) && cstack[op1].val1 == 0) {
      cstack[top] = cstack[cstack[top].val2];
      return 1;
    }
    if (islit(cstack[op2].opcode) && cstack[op2].val1 == 0) {
      cstack[top] = cstack[cstack[top].val1];
      return 1;
    }
  } else if (op == SUB) {
    op2 = cstack[top].val2;

    if (islit(cstack[op2].opcode) && cstack[op2].val1 == 0) {
      cstack[top] = cstack[cstack[top].val1];
      return 1;
    }
  } else if (op == MUL) {
    op1 = cstack[top].val1;
    op2 = cstack[top].val2;

    if (islit(cstack[op1].opcode)) {
      if (cstack[op1].val1 == 0) {
	cstack[top].opcode  = LIT;
	cstack[top].val1 = 0;
	return 1;
      } else if (cstack[op1].val1 == 1) {
	cstack[top] = cstack[cstack[top].val2];
	return 1;
      }

    } else if (islit(cstack[op2].opcode)) {
      if (cstack[op2].val1 == 0) {
	cstack[top].opcode  = LIT;
	cstack[top].val1 = 0;
	return 1;
      } else if (cstack[op2].val1 == 1) {
	cstack[top] = cstack[cstack[top].val1];
      return 1;
      }
    }
    return 0;
  }
}

doassoc(int top)
{
  int op;
  op = cstack[top].opcode;
  if (op == ADD) {
    if (islit(cstack[cstack[top].val1].opcode)) {
      if (cstack[cstack[top].val2].opcode == ADD) {
	if (islit(cstack[cstack[cstack[top].val2].val1].opcode)) {
	  cstack[cstack[top].val1].val1 = cstack[cstack[top].val1].val1 + cstack[cstack[cstack[top].val2].val1].val1;
	  cstack[top].val2 = cstack[cstack[top].val2].val2;
	} else if (islit(cstack[cstack[cstack[top].val2].val2].opcode)) {
	  cstack[cstack[top].val1].val1 = cstack[cstack[top].val1].val1 + cstack[cstack[cstack[top].val2].val2].val1;
	  cstack[top].val2 = cstack[cstack[top].val2].val1;
	}
      }
    } else if (islit(cstack[cstack[top].val2].opcode)) {
      if (cstack[cstack[top].val1].opcode == ADD) {
	if (islit(cstack[cstack[cstack[top].val1].val1].opcode)) {
	  cstack[cstack[top].val2].val1 = cstack[cstack[top].val2].val1 + cstack[cstack[cstack[top].val1].val1].val1;
	  cstack[top].val1 = cstack[cstack[top].val1].val2;
	} else if (islit(cstack[cstack[cstack[top].val1].val2].opcode)) {
	  cstack[cstack[top].val2].val1 = cstack[cstack[top].val2].val1 + cstack[cstack[cstack[top].val1].val2].val1;
	  cstack[top].val1 = cstack[cstack[top].val1].val1;
	}
      }
    }
  } else if (op == MUL) {
    if (islit(cstack[cstack[top].val1].opcode)) {
      if (cstack[cstack[top].val2].opcode == MUL) {
	if (islit(cstack[cstack[cstack[top].val2].val1].opcode)) {
	  cstack[cstack[top].val1].val1 = cstack[cstack[top].val1].val1 * cstack[cstack[cstack[top].val2].val1].val1;
	  cstack[top].val2 = cstack[cstack[top].val2].val2;
	} else if (islit(cstack[cstack[cstack[top].val2].val2].opcode)) {
	  cstack[cstack[top].val1].val1 = cstack[cstack[top].val1].val1 * cstack[cstack[cstack[top].val2].val2].val1;
	  cstack[top].val2 = cstack[cstack[top].val2].val2;
	}
      }
    } else if (islit(cstack[cstack[top].val2].opcode)) {
      if (cstack[cstack[top].val1].opcode == MUL) {
	if (islit(cstack[cstack[cstack[top].val1].val1].opcode)) {
	  cstack[cstack[top].val2].val1 = cstack[cstack[top].val2].val1 * cstack[cstack[cstack[top].val1].val1].val1;
	  cstack[top].val1 = cstack[cstack[top].val1].val2;
	} else if (islit(cstack[cstack[cstack[top].val1].val2].opcode)) {
	  cstack[cstack[top].val2].val1 = cstack[cstack[top].val2].val1 * cstack[cstack[cstack[top].val1].val2].val1;
	  cstack[top].val1 = cstack[cstack[top].val1].val1;
	}
      }
    }
  }
}


branchopt(top)
{
  int cond;

  if (cstack[top].opcode == IF) {
    cond = cstack[top].val1;
    if (cstack[cond].opcode == LIT) {
      if (cstack[cond].val1)
	cstack[top] = cstack[cstack[cstack[top].val2].val1];
      else 
	cstack[top] = cstack[cstack[cstack[top].val2].val2];
    }
  }
}


codeopt()
{

}

