#include "vcc.h"
#include "ptopcode.h"

#define islit(op) ((op)==LIT)
#define isvar(op) ((op)==VAR)
#define isoperator(op) ((op) >= ADD && (op) <= DIV || (op) >= GREATER && (op) <= NOTEQUAL)
#define iscomoperator(op) ((op) == ADD || (op) == MUL || (op) == EQUAL || (op) == NOTEQUAL)
#define compconst(op) {cstack[top].val1 = cstack[left].val1 op cstack[right].val1;}


#define VNARYSIZE 64
#define istempvar(v) ((v) < 0)

static struct _valnum_t {
  int node;
  int qty;
  int op;
  int operand1;
  int operand2;
} *valnum;

static int *qtyvar;

static int nodetop;
static int qtytop;
static int valnumtop;
static int qtyvartop;

int setvalnum(int nodeno, int op, int op1, int op2)
{
  int i, r;

  valnum[valnumtop].node = nodeno;
  if ((r = valnumcalc(op, op1, op2)) != 0) {
    valnum[valnumtop].qty = r;
    valnum[valnumtop].op = -1;
  } else {
    valnum[valnumtop].qty = qtytop++;
    valnum[valnumtop].op = op;
    valnum[valnumtop].operand1 = op1;
    valnum[valnumtop].operand2 = op2;
  }
  return valnum[valnumtop++].qty;
}

int valnumcalc(int op, int val1, int val2)
{
  int i;

  for (i = valnumtop-1; i >0; i--) {
    if (valnum[i].op == op) {
      if (valnum[i].operand1 == val1 && valnum[i].operand2 == val2) {
	if (debug)
	  fprintf(stderr, "QTY %d found\n", valnum[i].qty);
	return valnum[i].qty;
      } else if (iscomoperator(op)) { // handling commutative operators
	if (valnum[i].operand1 == val2 && valnum[i].operand2 == val1) {
	  if (debug)
	    fprintf(stderr, "QTY %d found\n", valnum[i].qty);
	  return valnum[i].qty;
	}
      }
    }
  }
  return 0; //not found
}

int valnumadd(int nodeno)
{
  // handles LIT, VAR, and Arithmetic
  int op, o1, o2, o3;

  if (nodeno == 0) return 0;
  op = cstack[nodeno].opcode;
  if (islit(op) || isvar(op)) {
    return setvalnum(nodeno, op, cstack[nodeno].val1, 0);
  } else if (isoperator(cstack[nodeno].opcode)) {
    o1 = valuenum(cstack[nodeno].val1);
    o2 = valuenum(cstack[nodeno].val2);
    return setvalnum(nodeno, op, o1, o2);
  }
}

int valnumaddmov(int nodeno, int qty)
{
    if (qty == 0) {
      qty = qtytop++;
    } 
    valnum[valnumtop].node = nodeno;
    valnum[valnumtop].qty = qty;
    valnum[valnumtop].op = VAR;
    valnum[valnumtop].operand1 = cstack[nodeno].val1;
    valnum[valnumtop].operand2 = 0;
    valnumtop++;
    return qty;
}

void valnuminit(int top)
{
  nodetop = top;
  qtytop = 1;
  valnumtop = 1;

}

char *ivarname(int i)
{
  static char buf[8];
  static char *smallname[5] = {"%0", "%1", "%2", "%3", "%4"};

  if (i < 5) return smallname[i];
  sprintf(buf, "%%%d", i);
  return buf;
}

int register_name(char *s)
{
  ENTRY e, *r;
  int pos;

  strcpy(namepool+namepooltop, s);
  e.key = namepool+namepooltop;
  hsearch_r(e, FIND, &r, &namhash);
  if (r == NULL) { // not found.
    e.data = (void *)namepooltop;
    hsearch_r(e, ENTER, &r, &namhash);
    namepooltop += strlen(s)+1;
  } 
  pos = (int)(r->data);

  return pos;
}

int isuseful(int start, int qty)
{
  int i, count, pos;

  if (qty == 0 || valnum[start].qty == 0) 
    return 0;

  if (valnum[start].op == -1)
      return 2;

  for (i = start+1, count = 0; i < valnumtop; i++) {
    if (valnum[i].qty == qty)
      return 1;
  }
  return 0;
}

void pushqty(int qty)
{
  int i;

  for (i = 0; i < qtyvartop; i++) {
    if (qtyvar[i] == qty) 
      break;
  }
  if (i == qtyvartop) 
    qtyvar[qtyvartop++] = qty;
}

void insertmov(int pos)
{
  int lhs, node, newnode;;

  node = valnum[pos].node;
  lhs = push(VAR, register_name(ivarname(valnum[pos].qty)), 0);

  pushqty(valnum[pos].qty);

  // swap node and newnode
  newnode = push(cstack[node].opcode, cstack[node].val1, cstack[node].val2);
  cstack[node].opcode = MOV;
  cstack[node].val1 = lhs;
  cstack[node].val2 = newnode;
}

void replacevar(int pos)
{
  if (isvar(cstack[valnum[pos].node].opcode))
    return;
  if (debug)
    fprintf(stderr, "replacevar *** with %d\n", valnum[pos].qty);
  cstack[valnum[pos].node].opcode = VAR;
  cstack[valnum[pos].node].val1 = register_name(ivarname(valnum[pos].qty));
  cstack[valnum[pos].node].val2 = 0;
}

void valnumrepl()
{
  int i;

  for (i =1; i < valnumtop; i++) {
    if (islit(cstack[valnum[i].node].opcode) || isvar(cstack[valnum[i].node].opcode)) {
     continue;
    }
    switch (isuseful(i, valnum[i].qty)) {
    case 1: 
      insertmov(i);
      break;
    case 2:
      replacevar(i);
      break;
    }
  }
}

void valnumreset()
{
  valnumrepl();
  qtytop = 1;
  valnumtop = 1;
}


int valuenum(int top)
{
  int op, o1, o2, savedtop;

  op = cstack[top].opcode;
  if (isoperator(op) || islit(op) || isvar(op)) {
    return valnumadd(top);
  } else {
    switch (op) {
    case MOV: 
      o1 = valuenum(cstack[top].val2);
      if (!isvar(cstack[cstack[top].val1].opcode)) {
	valuenum(cstack[top].val1);
	return 0;
      }
      o2 = valnumcalc(VAR, cstack[top].val1, 0);
      if (o1 > 0 && o1 == o2) {
	cstack[top].opcode = -1; // invalidate
	return o1;
      } else {
	return valnumaddmov(cstack[top].val1, o1);
      }
      break;
    case LISTREF:
      valuenum(cstack[top].val2);
      break;
    case FUNCALL:
      valuenum(cstack[top].val2); 
      valnumreset();
      break;
    case RET:
      valuenum(cstack[top].val1);
      valnumreset();
      break;
    case LENV:
      if (cstack[cstack[cstack[top].val1].val1].opcode == MYVAR) {
	valnumreset();
	valuenum(cstack[cstack[top].val1].val2);
      } else {
	valuenum(cstack[top].val1);
      }
      break;
    case IF: 
      valuenum(cstack[top].val1);
      savedtop = valnumtop;
      valuenum(cstack[cstack[top].val2].val1);
      valnumrepl();
      valnumtop = savedtop; 	// we need rollback
      valuenum(cstack[cstack[top].val2].val2);
      valnumreset();
      break;
    case COMPOUND:
      valuenum(cstack[top].val1);
      valuenum(cstack[top].val2);
      break;
    case WHILE:
      valnumreset();
      valuenum(cstack[top].val1);
      valnumreset();
      valuenum(cstack[top].val2);
      valnumreset();
      break;
    case ARG:
      valuenum(cstack[top].val1);
      valuenum(cstack[top].val2);
      break;
    default:
      break;
    }
    return 0;
  }
}

void addlvars(int top)
{
  int i, vlist, node, node1, newtop, res;

  if (qtyvartop == 0)
    return;

  newtop = push(0, 0, 0);
  cstack[newtop] = cstack[top];
  vlist = push(MYVAR, node = push(PARAM, 0, 0), 0);
  for (i = 0; i < qtyvartop; i++) {
    node1 = push(VAR, register_name(ivarname(qtyvar[i])), 0);
    cstack[node].val1 = node1;
    node1 = push(PARAM, 0, 0);
    cstack[node].val2 = node1;
    node = node1;
  } 
  cstack[top].opcode = LENV;
  cstack[top].val1 = push(COMPOUND, vlist, newtop);
  cstack[top].val2 = 0;
}


void valuenumbering(int fbody)
{
  valnum = malloc(sizeof(struct _valnum_t) * VNARYSIZE);
  qtyvar = malloc(sizeof(int) * VNARYSIZE);
  qtyvartop = 0;

  valnuminit(fbody);
  valuenum(fbody);
  valnumreset();
  addlvars(fbody);

  free(qtyvar);
  free(valnum);
}
