#include "vm.h"
#include "opcode.h"

#define CODESEGSIZE 1000
#define STACKLEN 1000
#define DATASEGSIZE 100000

#define HSIZE 257

extern int debug;

t_codeseg  codeseg[CODESEGSIZE];
t_stackseg stackseg[STACKLEN];
int dataseg[DATASEGSIZE];

t_dataseg  objectab[128];
int objectabtop;

int framestack[STACKLEN];

int codepc;
int stacktop;
int dsegtop;

int savedframetop;

struct hsearch_data otabhash;

#define binop(OP) {stackseg[stacktop-1] = (stackseg[stacktop-1] OP stackseg[stacktop]);stacktop--; }

#define NEXTPC {pc++;break;}

#define condjmp(OP) {if (stackseg[stacktop] OP) { pc += codeseg[pc].operand; stacktop--; break;} else { NEXTPC;}}

vminit()
{
  objectabtop = 0;
  savedframetop = 0;
  dsegtop = 0;

  if (hcreate_r(HSIZE, &otabhash) == 0) {
    fprintf(stderr, "hash memory for VM is not enough\n");
    exit(1);
  }
  listallocregister();
  createobjregister();
}

vmglobalvarsearch(int varname)
{
  ENTRY e, *r;
  e.key = namepool+varname;
  hsearch_r(e, FIND, &r, &otabhash);
  if (r == NULL) {
    int j;

    j = objectabtop++;
    objectab[j].typ = VAL;
    objectab[j].val = 0;
    objectab[j].paramlen = 0;

    e.data = (void *)j;
    hsearch_r(e, ENTER, &r, &otabhash);
    return j;
  } else {
    return (int)(r->data);
  }
}

clsmethodsearch(int cname, int mname)
{
  ENTRY e, *r;
  e.key = namepool+cname;
  int gv, i, plen, clsbody;

  hsearch_r(e, FIND, &r, &otabhash);
  if (r == NULL) {
    int j;

    return -1;
  } else {
    gv = (int)(r->data);
  }

  plen = objectab[gv].paramlen;
  clsbody = objectab[gv].val;
  for (i = 0; i < plen; i++) {
    if (methname(clsbody, i) == mname)
      return i;
  }
  if (debug) 
    printf("%s not found for %s\n", namepool+mname, namepool+cname);
  return -1;
}

vmmethodsearch(int obj, int meth)
{
  int clsname = objname(obj); 
  int gv = vmglobalvarsearch(clsname);
  int clsbody = (int)objectab[gv].val;
  int numethds = objectab[gv].paramlen;

  int i;
#if 0
  if (debug) 
    printf("obj = %s,  = %d\n", namepool+clsname, numethds);
#endif
  for (i = 0; i< numethds; i++) {
    if (methname(clsbody, i) == meth)
      return i;
  }
  return -1; /* field 0 is the name of the class */
}




void *cfunsearch();

vmexec(int pc)
{
  for (;;) {
    switch (codeseg[pc].op) {
    case OP_NOP:
      NEXTPC;
    case OP_POP:
      stacktop -= codeseg[pc].operand; NEXTPC;
    case OP_U_NOT:
      stackseg[stacktop] = !stackseg[stacktop]; NEXTPC;
    case OP_B_ADD:
      binop(+); NEXTPC;
    case OP_B_SUB:
      binop(-); NEXTPC;
    case OP_B_MUL:
      binop(*); NEXTPC;
    case OP_B_DIV:
      binop(/); NEXTPC;
    case OP_B_GREATEREQ:
      binop(>=); NEXTPC;
    case OP_B_GREATER:
      binop(>); NEXTPC;
    case OP_B_LESS:
      binop(<); NEXTPC;
    case OP_B_LESSEQ:
      binop(<=); NEXTPC;
    case OP_B_EQUAL:
      binop(==); NEXTPC;
    case OP_B_NOTEQUAL:
      binop(!=); NEXTPC;
    case OP_JMP:
      pc += codeseg[pc].operand; break;
    case OP_JMPPOS:
      condjmp(>0);
    case OP_JMPZ:
      condjmp(==0);
    case OP_JMPNEG:
      condjmp(<0);
    case OP_PUSH_CONST:
      stackseg[++stacktop] = codeseg[pc].operand; NEXTPC;
    case OP_PUSH_GVAR:
      {int gv = vmglobalvarsearch(codeseg[pc].operand);
      stackseg[++stacktop] = objectab[gv].val; NEXTPC;
      }
    case OP_PUSH_LVAR:
      stackseg[++stacktop] = stackseg[framestack[savedframetop-1]+codeseg[pc].operand]; NEXTPC;
    case OP_PUSH_ADDR:
      stackseg[stacktop] = dataseg[stackseg[stacktop]]; NEXTPC;
    case OP_PUSH_FIELD: { /* get field value */
      int m = vmmethodsearch(stackseg[stacktop], codeseg[pc].operand);

      stackseg[stacktop] = objmethod((stackseg[stacktop]), (m));
      NEXTPC;
    }
    case OP_DUP:
      stackseg[stacktop+1] = stackseg[stacktop]; stacktop++; NEXTPC;
    case OP_DUP2:
      stackseg[stacktop+1] = stackseg[stacktop-codeseg[pc].operand]; stacktop++; NEXTPC;
    case OP_STORE_GVAR:
      {int gv = vmglobalvarsearch(codeseg[pc].operand);
      objectab[gv].val = stackseg[stacktop--]; NEXTPC;
      }
    case OP_STORE_LVAR:
      stackseg[framestack[savedframetop-1]+codeseg[pc].operand] = stackseg[stacktop--]; NEXTPC;
    case OP_STORE_ADDR:
      dataseg[stackseg[stacktop]] = stackseg[stacktop-1]; stacktop-=1; NEXTPC;
    case OP_STORE_FIELD: {/* store field value */
      int m = vmmethodsearch(stackseg[stacktop], codeseg[pc].operand);

      objmethod(stackseg[stacktop], m) = stackseg[stacktop-1];
      stacktop-=2;
      NEXTPC;
    }
    case OP_CALL:
    case OP_CALL2: /* call2 for method invocation */
      { int cbody, ctyp, alen, gv;

	if (codeseg[pc].op == OP_CALL2) {
	  cbody = stackseg[stacktop--];
	  ctyp = FUN;
	} else { /* OP_CALL */

	  int r;

	  gv = vmglobalvarsearch(codeseg[pc].operand);
	  /* 	  alen = stackseg[stacktop--]; */
	  if (objectab[gv].val == 0) { /* undefined --> resolve */
	    void (*cfun)();

	    if ((cfun = cfunsearch(codeseg[pc].operand)) != NULL) {
	      objectab[gv].typ = CFUN;
	      objectab[gv].val = (void *)cfun;
	      objectab[gv].paramlen = 0; /* dummy */
	    }
	  }
	  cbody = objectab[gv].val;
	  ctyp = objectab[gv].typ;
	}
	alen = stackseg[stacktop--];
	switch (ctyp) { /* (objectab[gv].typ) { */
	case FUN:
	  framestack[savedframetop++] = pc+1;
	  /*	  framestack[savedframetop++] = stacktop-objectab[gv].paramlen+1; */
	  framestack[savedframetop++] = stacktop-alen+1;
	  pc = cbody; /* objectab[gv].val;  */
	  break;
	case CFUN:
	  {int r, (*f)() = objectab[gv].val;
	  r = f(alen, stackseg + (stacktop-alen+1));
	  stacktop -= alen;
	  stackseg[++stacktop] = r;
	  NEXTPC;
	  }
	default:
	  NEXTPC;
	}
	break;
      }
    case OP_RETURN:
      {
	int r = stackseg[stacktop]; 
#if 0
	if (debug)
	{int i;
	  for (i=0; i <= stacktop; i++)
	    printf("stack[%d] = %d::", i, stackseg[i]);
	  printf("\n");
	}
#endif
	stacktop = framestack[--savedframetop];
	stackseg[stacktop] = r;
	pc = framestack[--savedframetop];
	break;
      }
    case OP_PRINT_TOP:
      printf("%d\n", stackseg[stacktop--]);
      return;
    }
  }
}


listalloc(int ac, int av[])
{
  int r, i;

  r = dsegtop;
  dsegtop += av[0];
  if (ac > 1) {
    for (i = 0; i < av[0]; i++) {
      dataseg[r+i] = listalloc(ac-1, av+1);
    }
  }
  return r;
}

listalloc2(int size)
{
  int r;

  r = dsegtop;
  dsegtop+= size;
  return r;
}

extern int listkeywordpos;

listallocregister()
{
  int listallocpos;

  listallocpos = vmglobalvarsearch(listkeywordpos);

  objectab[listallocpos].typ = CFUN;
  objectab[listallocpos].val = (int)listalloc;
  objectab[listallocpos].paramlen = 2; /* dummy */
}

int createobj(int ac, int av[])
{
  int go = vmglobalvarsearch(av[0]);
  int r, size, i;

  if (debug)
    printf("creating %s\n", namepool+av[0]);
  if (objectab[go].typ != CLS) {
    fprintf(stderr, "Not a class name %s\n", namepool+av[0]);
    return 0;
  }

  size = objectab[go].paramlen+1;
  r = (int)listalloc2(size);

  objname(r) = av[0];
  for (i = 0; i < size; i++) {
    objmethod(r, i) = methbody(objectab[go].val, i);
  }
  return r;
}

createobjregister()
{
  int createobjpos;

  createobjpos = vmglobalvarsearch(newpos());

  objectab[createobjpos].typ = CFUN;
  objectab[createobjpos].val = (int)createobj;
  objectab[createobjpos].paramlen = 2; /* dummy */
}
